unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, MD2, CgWindow, CgUtils, GL, GLu, GLext, TGA2, Math, AppEvnts,
  ExtCtrls, StdCtrls, TypInfo;

type
  TMD2Form = class(TCGForm)
    AppEvents: TApplicationEvents;
    OpenDlg: TOpenDialog;
    pnlAnimControl: TPanel;
    lstAnimations: TListBox;
    chkLoopSelected: TCheckBox;
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure AppEventsIdle(Sender: TObject; var Done: Boolean);
    procedure chkLoopSelectedClick(Sender: TObject);
    procedure lstAnimationsClick(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

const
  NUM_FRAMES = 198;  // Number of keyframes in a Quake2 model.

var
  MD2Form: TMD2Form;
  time: Single = 0;           // Animation timer.
  vprog: GLuint = 0;          // Interpolation vertex program.
  skin: GLuint = 0;           // Model texture.
  anim_start: Integer = 0;    // First frame of animation sequence.
  anim_len: Integer = 198;    // Length of sequence (by default show all anims).
  ay: Single = 0;             // Y-axis rotation.
  mp: TPoint;                 // Mouse position.

// Type declarations for vertices:
type
  TVector = array [0..2] of Single;
  TTexCoord = array [0..1] of Single;
  TVertex = record
    Tex0: TTexCoord;
    Nrml: TVector;
    OPos: TVector;
  end;

var
  VertexArrays: array of array of TVertex; // One vertex array per keyframe.

implementation

{$R *.dfm}

// Load an MD2 model from file.
procedure LoadModel(const filename: String);
var
  theMD2: TMD2Model;
  f, i, j, v: Integer;
  tx, ty, tz, sx, sy, sz: Single;
  v0, v1, v2: TVertex;

  { This function turns a GLcommand into a full vertex that we can put in a
    vertex array. This is a brute-force approach that might duplicate a lot more
    vertices than strictly necessary. For best results, turn everything into an
    indexed triangle list so you can avoid duplicate vertices. }
  function Vertex(glc: TMD2GLCommand; frame: Integer): TVertex;
  var
    x, y, z: Single;
  begin
    Result.Tex0[0] := glc.s;
    Result.Tex0[1] := glc.t;

    with theMD2.Frames[frame].Vertices[glc.VertexIndex] do
    begin
      Result.Nrml[0] := MD2_VERTEX_NORMALS[LightNormalIndex][0];
      Result.Nrml[1] := MD2_VERTEX_NORMALS[LightNormalIndex][1];
      Result.Nrml[2] := MD2_VERTEX_NORMALS[LightNormalIndex][2];
    end;

    x := theMD2.Frames[frame].Vertices[glc.VertexIndex].Vertex[0];
    y := theMD2.Frames[frame].Vertices[glc.VertexIndex].Vertex[1];
    z := theMD2.Frames[frame].Vertices[glc.VertexIndex].Vertex[2];

    Result.OPos[0] := tx + x*sx;
    Result.OPos[1] := ty + y*sy;
    Result.OPos[2] := tz + z*sz;
  end;
begin

  // First load the MD2.
  theMD2 := TMD2Model.Create;
  theMD2.LoadFromFile(filename);

  if theMD2.Header.NumFrames <> 198 then 
    raise Exception.Create('This does not appear to be a standard Quake2 model!');

  // Now convert the MD2 keyframe data to a set of vertex arrays.
  SetLength(VertexArrays, theMD2.Header.NumFrames);

  for f := 0 to theMD2.Header.NumFrames - 1 do
  begin
    { Vertices are packed into bytes, but each frame has a scale and translation
      that expands the 3-byte vector back into a proper floating-point one. }
    sx := theMD2.Frames[f].Scale[0];
    sy := theMD2.Frames[f].Scale[1];
    sz := theMD2.Frames[f].Scale[2];

    tx := theMD2.Frames[f].Translate[0];
    ty := theMD2.Frames[f].Translate[1];
    tz := theMD2.Frames[f].Translate[2];

    SetLength(VertexArrays[f], 0);
    v := 0;

    for i := 0 to theMD2.Header.NumGLCommands - 1 do
    begin
      { Each GLcommand represents a strip or a fan, but we'll expand them all
        into a single big triangle list. }
      v0 := Vertex(theMD2.GLcommands[i].Commands[0], f);
      v1 := Vertex(theMD2.GLcommands[i].Commands[1], f);
      for j := 2 to High(theMD2.GLcommands[i].Commands) do
      begin
        v2 := Vertex(theMD2.GLcommands[i].Commands[j], f);

        SetLength(VertexArrays[f], (v+1) * 3);
        VertexArrays[f][v*3 + 0] := v0;
        VertexArrays[f][v*3 + 1] := v1;
        VertexArrays[f][v*3 + 2] := v2;

        if theMD2.GLCommands[i].Mode = MD2_GLCOMMAND_STRIP then v0 := v1;
        v1 := v2;
        INC(v);
      end;
    end;
  end;

end;

procedure LoadSkin(const filename: String);
var
  tgaskin: TTGAImage;
begin

  // Load the model's skin.
  tgaskin := TTGAImage.Create;
  tgaskin.LoadFromFile(filename);
  tgaskin.ExpandToTrueColor;
  glGenTextures(1, @skin);
  glBindTexture(GL_TEXTURE_2D, skin);
  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB8, tgaskin.Header.ImgSpec.Width,
                    tgaskin.Header.ImgSpec.Height, GL_BGR, GL_UNSIGNED_BYTE,
                    tgaskin.Data);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glEnable(GL_TEXTURE_2D);
  tgaskin.Free;

end;

procedure LoadShader;
var
  txt: TStringList;
  err: String;
  i: GLint;
begin

  // Load the keyframe interpolation shader.
  txt := TStringList.Create;
  txt.LoadFromFile('.\animate.vp');

  glEnable(GL_VERTEX_PROGRAM_ARB);
  glGenProgramsARB(1, @vprog);
  glBindProgramARB(GL_VERTEX_PROGRAM_ARB, vprog);
  glProgramStringARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB,
                     Length(txt.Text), PChar(txt.Text));

  txt.Free;

  err := PChar(glGetString(GL_PROGRAM_ERROR_STRING_ARB));
  if err <> '' then
  begin
    glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, @i);
    raise Exception.Create('Shader program contains errors:' + #10#10 + err +
                           Format(' (pos %d)', [i]));
  end;

end;

procedure TMD2Form.FormCreate(Sender: TObject);
var
  namemodel, nameskin, nameanim: String;
  i: TMD2Animation;
  ti: PTypeInfo;
begin

  try
    Show;

    // Pop up open dialogs to select the model and the skin.
    OpenDlg.Filter := 'MD2 models|*.md2';
    OpenDlg.Title := 'Choose model';
    if OpenDlg.Execute then namemodel := OpenDlg.FileName
    else raise Exception.Create('No model loaded.');

    OpenDlg.Filter := 'TGA images|*.tga';
    OpenDlg.Title := 'Choose model skin';
    if OpenDlg.Execute then nameskin := OpenDlg.FileName
    else raise Exception.Create('No skin loaded.');

    SetCurrentDir(ExtractFilePath(Application.ExeName));

    InitGL;

    if not glext_LoadExtension('GL_ARB_multitexture') then
      raise Exception.Create('This demo requires GL_ARB_multitexture!');

    if not glext_LoadExtension('GL_ARB_vertex_program') then
      raise Exception.Create('This demo requires GL_ARB_vertex_program!');

    LoadModel(namemodel);
    LoadSkin(nameskin);
    LoadShader;

    // Enable vertex arrays.
    glClientActiveTextureARB(GL_TEXTURE1_ARB);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTextureARB(GL_TEXTURE0_ARB);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_VERTEX_ARRAY);

    glEnable(GL_DEPTH_TEST);
    glClearColor(0.75, 0.75, 0.75, 1);

    // Fill the list box with the names of the animations.
    ti := TypeInfo(TMD2Animation);
    for i := Low(TMD2Animation) to High(TMD2Animation) do
    begin
      nameanim := GetEnumName(ti, Ord(i));
      Delete(nameanim, 1, 9);
      lstAnimations.Items.Add(nameanim);
    end;
    lstAnimations.ItemIndex := 0;

    GetCursorPos(mp);
    mp := ScreenToClient(mp);

    OnResize := FormResize;
    OnPaint := FormPaint;
    OnMouseMove := FormMouseMove;
    AppEvents.OnIdle := AppEventsIdle;
    // Force FormResize() to be called, so perspective gets set up.
    Resize;
    cgStartTiming;
  except on E: Exception do
    begin
      MessageDlg(E.Message, mtError, [mbOK], 0);
      Halt(1);
    end;
  end;

end;

procedure TMD2Form.FormResize(Sender: TObject);
begin

  glViewport(0, 0, ClientWidth - pnlAnimControl.Width, ClientHeight);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(60, (ClientWidth - pnlAnimControl.Width)/ClientHeight, 10, 1000);
  glMatrixMode(GL_MODELVIEW);

end;

procedure TMD2Form.FormPaint(Sender: TObject);
var
  f0, f1: Integer;
  lerp: Single;
  i: TMD2Animation;
  last: Boolean;
begin

  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  glLoadIdentity;

  glTranslatef(0, 0, -80);
  glRotatef(-90 + ay, 0, 1, 0);
  glRotatef(-90, 1, 0, 0); 

  // Determine the two frames we need to interpolate between...
  f0 := anim_start + (Floor(time) mod anim_len);
  f1 := anim_start + (Ceil(time) mod anim_len);
  // ... And also the interpolation weight.
  lerp := Frac(time);

  { If we're not looping a single animation, update the selection to highlight
    the animation that's currently playing. }
  if not chkLoopSelected.Checked then
  begin
    last := TRUE;
    for i := Low(TMD2Animation) to High(TMD2Animation) do
    begin
      if MD2_ANIM_FRAME_NUMBERS[i, 0] > f0 then
      begin
        lstAnimations.ItemIndex := Ord(i) - 1;
        last := FALSE;
        Break;
      end;
    end;
    if last then lstAnimations.ItemIndex := Ord(High(TMD2Animation));
  end;

  // Feed the interpolation weight to the vertex program.
  glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, lerp, lerp, lerp, 1);

  { We'll store the first of the two keyframes in the vertex position, and the
    second in the secondary texture coordinates. } 
  glClientActiveTextureARB(GL_TEXTURE1_ARB);
  glTexCoordPointer(3, GL_FLOAT, SizeOf(TVertex), @VertexArrays[f1][0].OPos);
  glClientActiveTextureARB(GL_TEXTURE0_ARB);
  glTexCoordPointer(2, GL_FLOAT, SizeOf(TVertex), @VertexArrays[f0][0].Tex0);
  glNormalPointer(GL_FLOAT, SizeOf(TVertex), @VertexArrays[f0][0].Nrml);
  glVertexPointer(3, GL_FLOAT, SizeOf(TVertex), @VertexArrays[f0][0].OPos);

  // Render!
  glDrawArrays(GL_TRIANGLES, 0, Length(VertexArrays[f0]));

  PageFlip;

end;

var
  fps: Single = 0;    // Sum of framerate samples.
  fpsc: Integer = 0;  // Number of framerate samples taken.

const
  FPS_SMOOVE = 25;    // Number of samples to use for smoothing the fps counter.

procedure TMD2Form.AppEventsIdle(Sender: TObject; var Done: Boolean);
var
  dt: Single;
  f: Integer;
begin

  Done := FALSE;

  dt := cgTimeElapsed;
  cgStartTiming;

  // How much time is there supposed to be between keyframes?
  time := time + dt/150;
  if time > 198 then time := time - 198;

  f := anim_start + (Floor(time) mod anim_len);

  // Take framerate sample.
  fps := fps + 1000/dt;
  INC(fpsc);

  // If number of samples is high enough, average them and display the result.
  if fpsc = FPS_SMOOVE then
  begin
    fpsc := 0;
    fps := fps / FPS_SMOOVE;
    Caption := Format('Quake2 MD2 Viewer -- %.0f fps -- frame %d/%d',
                      [fps, f, NUM_FRAMES]);
    fps := 0;
  end;

  Paint;

end;

procedure TMD2Form.chkLoopSelectedClick(Sender: TObject);
var
  i: Integer;
begin

  // Change the currently playing animation.
  if chkLoopSelected.Checked then
  begin
    i := lstAnimations.ItemIndex;
    anim_start := MD2_ANIM_FRAME_NUMBERS[TMD2Animation(i), 0];
    anim_len := MD2_ANIM_FRAME_NUMBERS[TMD2Animation(i), 1];
  end
  else begin
    anim_start := 0;
    anim_len := 198;
  end;

end;

procedure TMD2Form.lstAnimationsClick(Sender: TObject);
var
  i: Integer;
begin

  // Change the currently playing animation.
  if chkLoopSelected.Checked then
  begin
    i := lstAnimations.ItemIndex;
    anim_start := MD2_ANIM_FRAME_NUMBERS[TMD2Animation(i), 0];
    anim_len := MD2_ANIM_FRAME_NUMBERS[TMD2Animation(i), 1];
  end;

end;

procedure TMD2Form.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin

  // Rotate the model around the Y axis.
  if ssLeft in Shift then
  begin
    ay := ay + (X - mp.X);
    if ay > 360 then ay := ay - 360
    else if ay < 0 then ay := ay + 360;
  end;
  mp := Point(X, Y);

end;

end.

